use crate::prelude::*;
use anyhow::Result;
use assetpath::AssetPath;
use utilities::prelude::*;

use std::sync::{
    atomic::{AtomicBool, Ordering::SeqCst},
    Arc,
};

#[derive(Debug, Clone, PartialEq)]
pub enum FillTypeInfo {
    Image(AssetPath),
    Color(Color),
}

impl From<Color> for FillTypeInfo {
    fn from(color: Color) -> Self {
        Self::Color(color)
    }
}

impl From<String> for FillTypeInfo {
    fn from(s: String) -> Self {
        Self::Image(AssetPath::from(s))
    }
}

impl From<&str> for FillTypeInfo {
    fn from(s: &str) -> Self {
        Self::Image(AssetPath::from(s))
    }
}

impl From<&String> for FillTypeInfo {
    fn from(s: &String) -> Self {
        Self::Image(AssetPath::from(s.as_str()))
    }
}

impl From<AssetPath> for FillTypeInfo {
    fn from(assetpath: AssetPath) -> Self {
        Self::Image(assetpath)
    }
}

pub(crate) enum InnerFillType {
    Image(Arc<Displayable>),
    Color(Arc<Colorable>),
}

impl InnerFillType {
    pub(crate) fn new(framable: Arc<Framable>, fill_type_info: FillTypeInfo) -> Result<Self> {
        match fill_type_info {
            FillTypeInfo::Image(path) => {
                Ok(InnerFillType::Image(Displayable::new(framable, path)?))
            }
            FillTypeInfo::Color(color) => {
                Ok(InnerFillType::Color(Colorable::new(framable, color)?))
            }
        }
    }

    pub(crate) fn enable(&self) -> Result<()> {
        match self {
            Self::Image(displayable) => displayable.add()?,
            Self::Color(colorable) => colorable.add()?,
        }

        Ok(())
    }

    pub(crate) fn disable(&self) -> Result<()> {
        match self {
            Self::Image(displayable) => displayable.delete()?,
            Self::Color(colorable) => colorable.delete()?,
        }

        Ok(())
    }

    pub(crate) fn update_frame(&self) -> Result<()> {
        match self {
            Self::Image(displayable) => displayable.update_frame()?,
            Self::Color(colorable) => colorable.update_frame()?,
        }

        Ok(())
    }
}

impl Drop for InnerFillType {
    fn drop(&mut self) {
        match self {
            Self::Image(displayable) => displayable.clear_callback(),
            Self::Color(colorable) => colorable.clear_callback(),
        }
    }
}

pub(crate) struct FillType {
    pub(crate) inner: InnerFillType,
    visible: AtomicBool,
}

impl FillType {
    pub(crate) fn new(framable: Arc<Framable>, fill_type_info: FillTypeInfo) -> Result<Self> {
        let me = Self {
            inner: InnerFillType::new(framable.clone(), fill_type_info)?,
            visible: AtomicBool::new(false),
        };

        me.set_ui_layer(framable.ui_layer());

        Ok(me)
    }

    pub(crate) fn enable(&self) -> Result<()> {
        if !self.visible.load(SeqCst) {
            self.visible.store(true, SeqCst);
            self.inner.enable()?;
        }

        Ok(())
    }

    pub(crate) fn disable(&self) -> Result<()> {
        if self.visible.load(SeqCst) {
            self.visible.store(false, SeqCst);
            self.inner.disable()?;
        }

        Ok(())
    }

    pub(crate) fn update_frame(&self) -> Result<()> {
        self.inner.update_frame()
    }

    pub(crate) fn set_ui_layer(&self, layer: i32) {
        match &self.inner {
            InnerFillType::Image(displayable) => displayable.set_ui_layer(layer),
            InnerFillType::Color(colorable) => colorable.set_ui_layer(layer),
        }
    }
}

impl Drop for FillType {
    fn drop(&mut self) {
        if self.visible.load(SeqCst) {
            match &self.inner {
                InnerFillType::Image(displayable) => displayable.delete().unwrap(),
                InnerFillType::Color(colorable) => colorable.delete().unwrap(),
            }
        }
    }
}
